/***********************************************************************
|  Genealogy management system
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* io.c: supporting I/O routines for genealogy database */

#include "structs.h"

getletter(s)
char *s;
{
int c;
while(1) {
c = get_no_echo();
if(c == 'q') { nl(); exit(0); }
if(c == '.') longjmp(main_menu, 2);
if(strchr(s, c)) return c;
bell();
}
} /* getletter */

promptfor(msg, s)
char *msg, *s;
{
printf("%s ", msg);
return getletter(s);
} /* promptfor */

get_yn(msg)
char *msg;
{
char c = promptfor(msg, "nNyY");
nl();
return c&1;
} /* get_yn */

clearscreen()
{
nl();
} /* clearscreen */

db_is_empty()
{
struct rel_cursor c;
struct person *p;

LOOP_person(&c, p, NULL)
break;
END_LOOP_person(&c)
return p ? 0 : 1;
} /* db_is_empty */

set_curperson(l)
long l;
{
char *sn;

if(!select_copy_person(l, &cur_prs))
eprint("cannot read person record for serial %ld", l);
cur_prs_rowid = l;
birthname(l);
strcpy(cur_fname, fld1);
strcpy(cur_mname, fld2);
strcpy(cur_lname, fld3);

cur_birthplace[0] = 0;
sn = shortnote(cur_prs_rowid, 'b');
if(sn) strcpy(cur_birthplace, sn);
} /* set_curperson */

birthname(prsid)
long prsid;
{
static struct name n, *n1;
struct rel_cursor c;

n.birth = 0;
IDX_LOOP_name(&c, n1, &n, prsid, "prsid")
birth_rowid = c.rowid;
if(n.birth) break;
END_LOOP_name(&c)

if(!n.birth)
eprint("no birth name for person %ld", prsid);
splitstring(n.fullname);
} /* birthname */

fullbirthname(prsid, s)
long prsid;
char *s;
{
birthname(prsid);
strcpy(s, fld1);
strcat(s, " ");
if(fld2[0]) {
strcat(s, fld2);
strcat(s, " ");
}
strcat(s, fld3);
} /* fullbirthname */

/* retrieve a short note */
char *shortnote(prsid, ntype)
long prsid;
char ntype;
{
struct rel_cursor c;
static struct qknote qn, *q1;

IDX_LOOP_qknote(&c, q1, &qn, prsid, "prsid")
birth_rowid = c.rowid;
if(qn.ntype == ntype) break;
END_LOOP_qknote(&c)

if(q1) return qn.note;
return NULL;
} /* shortnote */

date getdate(datename, reason, defyear)
char *datename;
char *reason;
{
int m, d, y;
long l;

top:
getline(datename, "date", "#/", "digits and spaces", defyear ? 2 : 0);
if(!splitstring(line)) {
if(!defyear) return 0L;
printf("Enter a month and a day, e.g. `5 31' or `7/13'\n");
goto top;
}

y = 0;
if(fld3[0]) y = atoi(fld3);
if(!y) y = defyear;
if(defyear && y != defyear) {
printf("year %d is not the same as the established year %d\n", y, defyear);
goto top;
}
if(!y) {
printf("enter three numbers indicating month, day, and year:   e.g. 3 14 1960\n");
printf("Specify at least the year, month and day are optional\n");
goto top;
}
if(y >= 2500 || y < 200) {
printf("year %d out of range; years must lie between 1000 and 2500\n", y);
goto top;
}

/* now for the month and day */
if(!fld1[0]) { /* month/day not specified */
l = YEAR_DAY_RATIO * y;
} else {

m = atoi(fld1);
if(!fld2[0]) d = 1;
else d = atoi(fld2);
l = datecv(m, d, y);
if(!l) goto top;
}

if(lowdate) {
if(l < lowdate || l > highdate) {
printf("%s\n", reason);
goto top;
}
}

return l;
} /* getdate */

char *qkdesc(ntype, longform)
{
static char choices[] = "bdeorma";
static char *desc[2][7] = {
{"Born", "Died", "Educ", "Occup", "Relig", "Medic", "Ances"},
{"birthplace", "place and/or circumstances of death",
"education", "occupation", "religion",
"medical conditions", "country or ethnicity of recent ancestors"}
};
int ind;

ind = strchr(choices, ntype) - choices;
return desc[longform][ind];
} /* qkdesc */

getqknote(ntype)
char ntype;
{
char *prompt = qkdesc(ntype, 1);

printf("Enter %s\n", prompt);
while(1) {
printf("Info : ");
gets(line);
if(!strcmp(line, ".")) longjmp(main_menu, 1);
if(!strcmp(line, "q")) exit(0);
if(strlen(line) <= 72) break;
printf("Shortnote too long, please limit to 72 characters\n");
}
} /* getqknote */

lifetime_dates(p, fname, sloplow, slophigh)
struct person *p;
char *fname;
{
lowdate = highdate = 0;
if(p->birth) {
lowdate = p->birth + sloplow*YEAR_DAY_RATIO;
highdate = p->birth + slophigh*YEAR_DAY_RATIO;
sprintf(datemsg, "date is out of range relative to %s's lifetime", fname);
if(p->death && p->death < highdate)
highdate = p->death;
if(sloplow >= 0)
sprintf(datemsg, "date is not within %s's lifetime", fname);
}
} /* lifetime_dates */

linktype_dates(p, fname, linktype, adopt)
struct person *p;
char *fname;
char linktype;
{
int minage;

/* the minimum "reasonable" age to have a child is 10.
 * But if the child is adopted, the age difference could be less.
 * A 18 year old could marry somebody with a 12-year-old child. */
minage = adopt ? 3 : 10;

switch(linktype) {
case 'p':
lowdate = highdate = 0;
break;
case 'f': case 'm':
lifetime_dates(p, fname, -80, -minage);
break;
case 'c':
lifetime_dates(p, fname, minage, 80);
break;
case 'h': case 'w':
lifetime_dates(p, fname, -80, 80);
break;
default:
eprint("bad link type %c in linktype_dates()", linktype);
} /* end switch on link type */
} /* linktype_dates */

/* read input line and normalize */
/* do not call this with any open cursors, it may longjmp back to start. */
getline(prompt, item, lineletters, lineformdesc, deflast)
char *prompt, *item;
char *lineletters, *lineformdesc;
{
char space, c, badchar, specchar, atmode, sermode;
int i, j;
char *s, *t;

top:
printf("Enter %s: ", prompt);
if(gets(line) == 0) exit(0); /* EOF */
if(line[1] == 0) {
if(line[0] == '.') longjmp(main_menu, 1);
if(line[0] == 'q') exit(0);
}

sermode = atmode = 0;
specchar = 0;

/* compress white space */
space = 1;
for(i=j=0; c = line[i]; ++i) {
if(isspace(c)) {
if(!space) {
line[j++] = ' ';
space = 1;
}
continue;
}

space = 0;
line[j++] = c;
badchar = 1;

if(isalpha(c) && strchr(lineletters, 'L'))
badchar = 0;
if(isdigit(c) && strchr(lineletters, '#'))
badchar = 0;
if(strchr(lineletters, c) && strchr("/-'", c)) {
if(j > 1 && line[j-2] == ' ') line[--j-1] = c;
space = 1;
badchar = 0;
if(c == '/') ++specchar;
}
if(strchr("*@#",c) && j == 1 && strchr(lineletters, '@')) {
deflast = 0;
space = 1;
badchar = 0;
if(c == '@') {
if(!cur_prs_rowid) {
printf("Cannot use @ directive, no current person established.\n");
goto top;
}
atmode = 1;
continue;
}
if(c == '#') { sermode = 1; continue; }
}

if(atmode) {
space = 1;
badchar = 0;
if(!strchr(">hwfmcoy123456789",c)) {
printf("Bad character %c in @ directive.\n\
Use only fmchwyo> or 1-9.\n", c);
goto top;
}
}

if(!badchar) continue;

printf("%s contains invalid character %c, use only %s\n",
item, line[j-1], lineformdesc);
goto top;
} /* end loop */

if(specchar && specchar != 2) {
printf("please use exactly two slashes / to separate the three fields\n");
goto top;
}

/* get rid of last space */
if(j && line[j-1] == ' ') --j;
line[j] = 0;

if(atmode) return;

if(strchr(lineletters, '/') && !specchar) {
/* turn spaces into slashes */
s = strchr(line, ' ');
if(!s) {
if(deflast) {
strcat(line, "//");
if(deflast == 1 && strlen(line) > 2) {
smartmix(line, 0);
strcat(line, cur_lname);
}
} else {
shiftright(line,'/'); shiftright(line,'/');
}
} else {
*s = '/';
t = strchr(++s, ' ');
if(!t) {
if(deflast) {
strcat(line, "/");
if(deflast == 1) {
smartmix(line, 0);
strcat(line, cur_lname);
}
} else {
shiftright(line,'/');
*s = 0;
shiftleft(line);
*s = '/';
*--s = '/';
}
} else {
while(s = strchr(t+1, ' '))
t = s;
*t = '/';
}
}
} /* end slash normalization */
} /* getline */

struct name *getname(whose, nicksearch, deflast)
char *whose;
int deflast; /* use cur_lname as default */
{
static struct name n;
int bits;
static char lrs[5] = "-'L/";

lrs[4] = nicksearch ? '@' : 0;

top:
init_0(&n);
getline(whose, "name", lrs, "letters and spaces, or ' -", deflast);
if(line[0] == '@') {
shiftleft(line);
n.search = 2;
return &n;
}

if(strlen(line) >= NAMELEN) {
printf("name is too long, limit %d characters\n", NAMELEN-1);
goto top;
}
smartmix(line, 0);

bits = splitstring(line);
if((bits&5) != 5) {
if(!deflast)
printf("Enter at least a first and a last name, middle name is optional\n");
else
printf("Enter at least the first name, last name %s is assumed\n", cur_lname);
goto top;
} /* bad bits */

if(line[0] == '*') {
shiftleft(line);
strcpy(fld1, fld1+1);
n.search = 1;
}
strcpy(n.fullname, line);
gng_hashname(&n);
return &n;
} /* getname */

change_lname(n, lname)
struct name *n;
char *lname;
{
char *fullname = n->fullname;
char *t = strrchr(fullname, '/');
++t;
if(t-fullname + strlen(lname) >= NAMELEN) {
printf("adjusted name is too long, limit %d characters\n", NAMELEN-1);
longjmp(main_menu, 3);
}
strcpy(t, lname);
gng_hashname(n);
} /* change_lname */

/* separate / delimited fields in line[] and place them in fld[123].
 * Return bits indicating which fields are actually present. */
splitstring(s)
char *s;
{
char *t;
int n, fields = 0;

t = strchr(s, '/');
n = t-s;
strncpy(fld1, s, n);
fld1[n] = 0;
if(n) fields |= 1;

s = t+1;
t = strchr(s, '/');
n = t-s;
strncpy(fld2, s, n);
fld2[n] = 0;
if(n) fields |= 2;

s = t+1;
n = strlen(s);
strcpy(fld3, s);
if(n) fields |= 4;

return fields;
} /* splitstring */

gng_hashname(n)
struct name *n;
{
splitstring(n->fullname);
n->hashval = hashname(fld1, fld3);
} /* gng_hashname */

char *gng_nextfile()
{
char *fn, *t;

top:
fn = nextfile("text");
if(fn) {
if(!isdigit(fn[5])) goto top;
t = strchr(fn, '.');
if(!t || t[1] != 'Y') goto top;
}
return fn;
} /* gng_nextfile */

