/***********************************************************************
|  Basic calculator with polynomials
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* bccmd.c: bc keyword commands */

#include "bcgram.h"


/* keyword commands for bc */
bc_cmd(cmd, p) /*###*/
struct POLY *p;
{
register i;
struct VINT *t = 0;
struct VINT *t1;
struct POLY *q;
int n;
char first;
extern int scanner_base;

if(err_seen) return;

if(cmd >= TOK_LC_PRIME && !p) { eprint("no argument present"); return; }

if(p) {
if(isconst(p)) t = p->terms[0].coef;
if(cmd >= TOK_LC_SQRT && !t) { eprint("command requires a constant"); return; }
if(cmd == TOK_LC_PRIME || cmd == TOK_LC_FACTOR) {
n = n_vars(p);
if(n > 1) { eprint("polynomial must have one indeterminant"); return; }
if(n && (!modulus || !isprime(modulus))) {
eprint("polynomial must lie in a finite field of prime order, set modulus");
return;
}
}
}

switch(cmd)
{
case TOK_LC_QUIT: exit(0);

case TOK_LC_MOD: /* set modulus */
t1 = modulus;
modulus = &izero;
if(!t) break;
if(t->size < 0) { eprint("negative modulus\n"); break; }
modulus = t;
break;

case TOK_LC_IBASE: /* set base */
case TOK_LC_OBASE:
if(t) {
if(t->size != 1 || t -> value[0] > 36 || t -> value[0] < 2) {
eprint("invalid redix value\n");
break;
}
if(cmd == TOK_LC_IBASE) scanner_base = ibase.value[0] = t->value[0]; else obase.value[0] = t->value[0];
} else {
if(cmd == TOK_LC_IBASE) scanner_base = ibase.value[0] = 10; else obase.value[0] = 10;
}
break;

case TOK_LC_PRIME: /* prime test */
if(n) { eprint("not yet implemented"); break; }
/* remove minus sign */
t = iadd(t, &izero);
t->size &= 0x7fff;
if(!t->size || t->size == 1 && t->value[0] == 1) {
printf("not defined\n");
break;
}
if(isprime(t)) printf("prime\n");
else printf("composite\n");
break;

case TOK_LC_FACTOR: /* factor value */
if(n) { eprint("not yet implemented"); break; }
/* remove minus sign */
t = iadd(t, &izero);
t->size &= 0x7fff;
if(!t->size) { printf("0\n"); break; }

/* remove factors of 2 */
first = 1;
while(!(t->value[0] & 1)) {
if(first) first = 0; else printf("*");
ishr(t);
printf("2");
} /* factors of 2 */

if(t->size != 1 || t->value[0] != 1) {
if(!first) printf("*");
factor(t);
} else {
if(first) printf("1");
printf("\n");
}
break;

case TOK_LC_TOP: /* top */
q = padd(p, &pzero);
pdisp(q);
printf("\n");
break;

case TOK_LC_BOTTOM: /* bottom */
if(!p->next) pdisp(&pone);
else {
q = padd(p->next, &pzero);
pdisp(q);
}
printf("\n");
break;

case TOK_LC_SQRT: /* square root */
t = isqr(t);
q = newpoly(1);
q->terms[0].coef = t;
pdisp(q);
if(is_square) printf("*");
printf("\n");
break;

default:
eprint("@default[1] %d in bccmd.c", cmd);
} /* end switch on function */
} /* bc_cmd */

/* is polynomial a constant? */
isconst(p) /*###*/
struct POLY *p;
{
if(p->next) return 0;
if(p->nterms > 1) return 0;
if(p->terms[0].nvars > 0) return 0;
return 1;
} /* isconst */

/* number of variables in polynomial */
n_vars(p) /*###*/
struct POLY *p;
{
struct TERM *t;
register i;
int n = 0;
long list = 0L;

for(t=p->terms; t<p->terms+p->nterms; ++t)
for(i=0; i<t->nvars; ++i)
list |= (1L << t->vars[i]);
if(p->next)
{
p = p->next;
for(t=p->terms; t<p->terms+p->nterms; ++t)
for(i=0; i<t->nvars; ++i)
list |= (1L << t->vars[i]);
}

for(i=0; i<26; ++i, list>>=1)
n += list&1;

return n;
} /* n_vars */

