/***********************************************************************
|  Permutation group analyzer
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* groupio.c: get a Rubik cube permutation from stdin */

#include "grpnode.h"

static short tp1[PERMSIZE], tp2[PERMSIZE];
static short remap[PERMSIZE];
static char idncube[6][9] = {
{0,0,0,0,0,0,0,0,0},
{1,1,1,1,1,1,1,1,1},
{2,2,2,2,2,2,2,2,2},
{3,3,3,3,3,3,3,3,3},
{4,4,4,4,4,4,4,4,4},
{5,5,5,5,5,5,5,5,5}};

/* p is a pointer to an area that will hold the entered permutation.
*  type is a parameter specifying what cube information to expect
*  bits for type:
*  01  get corner pieces.
*  02  get side pieces.
*  04  get center pieces
*/
getcube(type,p)
short *p;
{
	static char *firstrow[]={
		"back","top","top","top","top"			,"front"};
	static char *facename[]={
"top","front","right","back","left","bottom"};
	char cube[6][9]; /* user-entered cube */
	/* number of colors per face (based on type) */
	static char cpf[]={		0,4,4,8,1,5,5,9			};
	int len=cpf[type];
	int cl=len/2; /* location of center square in input line */
	/* side adjacency tables */
	/* list of pairs, each pair making a side-piece */
	static char adjs[]={
1,28,3,37,5,19,7,10,
46,16,48,43,50,25,52,34,
12,41,14,21,23,30,32,39};
	/* corner adjacency tables */
	/* list of triples, each triple making a corner-piece */
	static char adjc[]={
0,36,29,2,27,20,6,9,38,8,18,11,
45,44,15,47,17,24,51,35,42,53,26,33};
int i,j;
	char s[60]; /* input line */
	int face;
	char temp[6];
	int nc; /* number of colors */
	char order[6]; /* colors for the 6 sides */
	char oppose[6]; /* which colors are opposite */
	char cmat[6][6]; /* color connection matrix */


top : 	nc=face=0;
nl();

	while(face<6){
		printf("Enter %s face, %s row first\n",facename[face],firstrow[face]);
		if(!gets(s)) exit(0);  /* EOF */

		/* discard white space */
		for(i=j=0;s[i];++i)
			if(s[i]!=' ' && s[i]!='\t') s[j++]=s[i];
		if(j!=len){
			fprintf(stderr,"enter %d colors, each color represented by a character of your choice\n",len);
			continue;
		}

		/* convert color characters into numbers */
		for(i=0;i<len;++i){
			for(j=0;j<nc;++j)
				if(temp[j]==s[i]) break;
			if(j==6){
				fprintf(stderr,"too many color characters, only use 6.\n\
\t(e.g. Y B O G R W.\n");
				goto top;
			}
			if(j==nc) temp[nc++]=s[i];
			cube[face][i]=j;
		}

		++face;
} /* end reading all six faces */

memset(order, 6, sizeof(order));

	/* check centers (if available) */
	if(type&4){
		for(i=0;i<6;++i) {
if(order[cube[i][cl]] != 6) {
					fprintf(stderr,"each center square must be a different color\n");
					goto top;
}
order[cube[i][cl]] = i;
if(type != 7)
memcpy(cube[i]+cl, cube[i]+cl+1, len-cl-1);
}
} /* center squares */

	/* move sides and corner numbers into standard positions */
for(i=0; i<6; ++i) {
if((type&3) == 1) {
cube[i][8] = cube[i][3];
cube[i][7] = 6;
cube[i][6] = cube[i][2];
cube[i][5] = 6;
cube[i][4] = 6;
cube[i][3] = 6;
cube[i][2] = cube[i][1];
cube[i][1] = 6;
}
if((type&3) == 2) {
cube[i][8] = 6;
cube[i][7] = cube[i][3];
cube[i][6] = 6;
cube[i][5] = cube[i][2];
cube[i][4] = 6;
cube[i][3] = cube[i][1];
cube[i][2] = 6;
cube[i][1] = cube[i][0];
cube[i][0] = 6;
}
} /* standard positions */

	/* color frequency */
	if(type&1){
memset(temp,0,sizeof(temp));
		for(i=0;i<6;++i)
			for(j=0;j<=8;j+=2)
if(j != 4)
				++temp[cube[i][j]];
		for(i=0;i<6;++i)
			if(temp[i]!=4){
				fprintf(stderr,"bad color distribution on the corners, each color should appear 4 times\n");
				goto top;
			}
	}

	if(type&2){
memset(temp,0,sizeof(temp));
		for(i=0;i<6;++i)
			for(j=1;j<=7;j+=2)
				++temp[cube[i][j]];
		for(i=0;i<6;++i)
			if(temp[i]!=4){
				fprintf(stderr,"bad color distribution on the sides, each color should appear 4 times\n");
				goto top;
			}
	}

	/* look at side color combinations */
	if(type&2){
memset(cmat, 0, sizeof(cmat));
		for(i=0;i<24;i+=2){
			char c1=cube[0][adjs[i]], c2=cube[0][adjs[i+1]];
			if(c1==c2){
				fprintf(stderr,"each side piece should consist of two different colors\n");
				goto top;
			}
			if(cmat[c1][c2]){
				fprintf(stderr,"each side piece should be different (a unique color pair)\n");
				goto top;
			}
			cmat[c1][c2]=cmat[c2][c1]=1;
		}
		for(i=0;i<5;++i)
			for(j=i+1;j<6;++j)
				if(!cmat[i][j]) oppose[i]=j,oppose[j]=i;
	}

	/* look at corner color combinations */
	if(type&1){
memset(cmat, 0, sizeof(cmat));
		for(i=0;i<24;i+=3){
			char c1=cube[0][adjc[i]], c2=cube[0][adjc[i+1]], c3=cube[0][adjc[i+2]];
			if(c1==c2 || c1==c3 || c2==c3){
				fprintf(stderr,"each corner should consist of 3 different colors\n");
				goto top;
			}
			if(cmat[c1][c2] | cmat[c2][c3] | cmat[c3][c1]){
				fprintf(stderr,"each corner piece should be different (a unique color triple)\n");
				goto top;
			}
			cmat[c1][c2]=cmat[c2][c3]=cmat[c3][c1]=1;
		}
		for(i=0;i<5;++i)
			for(j=i+1;j<6;++j){
				if(cmat[i][j]) continue;
				if((type&3)==3 && oppose[i]!=j){
					fprintf(stderr,"corners and sides produce different sets of oppositely colored faces\n");
					goto top;
				}
				oppose[i]=j,oppose[j]=i;
			}
	}

	/* do opposites match centers? */
	if(type&4){
		for(i=0;i<6; ++i) {
char f1 = order[i], f2 = order[oppose[i]], f3;
if(f2 < f1) f3 = f1, f1 = f2, f2 = f3;
if(f1 == 0 && f2 == 5) continue;
if(f1 == 1 && f2 == 3) continue;
if(f1 == 2 && f2 == 4) continue;
				fprintf(stderr,"pieces and centers produce different sets of oppositely colored faces\n");
				goto top;
			}
	}

	/* if no centers, make up a reasonable order[] */
	if((type&5)==1){
order[cube[0][51]] = 5;
order[cube[0][35]] = 3;
order[cube[0][42]] = 4;
order[oppose[cube[0][51]]] = 0;
order[oppose[cube[0][35]]] = 1;
order[oppose[cube[0][42]]] = 2;
	}

	if(type==2){
char f13 = 1;
order[cube[0][3]] = 0;
order[cube[0][37]] = 4;
order[oppose[cube[0][3]]] = 5;
order[oppose[cube[0][37]]] = 2;
		for(i=0;i<6;++i){
if(order[i] < 6) continue;
order[i] = f13;
f13 = 3;
}
} /* type = 2 */

	/* This does not quite guarantee that
	*  all pieces are really legal Rubik cublets,
	*  but it comes damn close, and i am lazy.
	*  all in all, this routine is a lot longer and more complex
	*  than i expected it to be. */

/* normalize colors using order[] */
	for(i=0;i<6;++i)
		for(j=0;j<=8;++j){
if(cube[i][j] == 6) continue;
				cube[i][j]=order[cube[i][j]];
		}

/* copy permutation into p */
permcpy(p, idn->p);
if(type&4)
for(i=0; i<6; ++i) p[i*9+4] = i*9+4;

if(type&1) {
for(i=0; i<24; i+=3) {
char c1 = cube[0][adjc[i]], c2 = cube[0][adjc[i+1]], c3 = cube[0][adjc[i+2]];
/* look through standard cube to find this color triple */
for(j=0; j<24; j+=3) {
char d1 = idncube[0][adjc[j]], d2 = idncube[0][adjc[j+1]], d3 = idncube[0][adjc[j+2]];
if(c1 == d1 && c2 == d2 && c3 == d3) {
p[adjc[j]] = adjc[i], p[adjc[j+1]] = adjc[i+1], p[adjc[j+2]] = adjc[i+2];
break;
}
if(c1 == d2 && c2 == d3 && c3 == d1) {
p[adjc[j+1]] = adjc[i], p[adjc[j+2]] = adjc[i+1], p[adjc[j]] = adjc[i+2];
break;
}
if(c1 == d3 && c2 == d1 && c3 == d2) {
p[adjc[j+2]] = adjc[i], p[adjc[j]] = adjc[i+1], p[adjc[j+1]] = adjc[i+2];
break;
}
}
}
}

if(type&2) {
for(i=0; i<24; i+=2) {
char c1 = cube[0][adjs[i]], c2 = cube[0][adjs[i+1]];
/* look through standard cube to find this color triple */
for(j=0; j<24; j+=2) {
char d1 = idncube[0][adjs[j]], d2 = idncube[0][adjs[j+1]];
if(c1 == d1 && c2 == d2) {
p[adjs[j]] = adjs[i], p[adjs[j+1]] = adjs[i+1];
break;
}
if(c1 == d2 && c2 == d1) {
p[adjs[j+1]] = adjs[i], p[adjs[j]] = adjs[i+1];
break;
}
}
}
}

remap_perm(p);
} /* getcube */

remap_set(p)
short *p;
{
remap_mode = 1;
permcpy(remap, p);
} /* remap_set */

/* remap a permutation.
 * This is rather like a base change in linear algebra.
 * Conjugate each group element by remap. */
remap_perm(p)
short *p;
{
if(!remap_mode || opt_q) return;
apply(remap, p, tp1, 0);
apply(tp1, remap, p, 1);
} /* remap_perm */

/* convert ascii string to a permutation */
line_to_perm(s, p, permsize)
char *s;
short *p;
{
int i = 0;
long strtol();

spacecrunch(s);
while(*s) {
p[i] = strtol(s, &s, 0);
if(p[i] < 0 || p[i] >= permsize) { i = 0; break; }
while(*s == ' ' || *s == ',') ++s;
if(++i == permsize) break;
}
if(i < permsize || *s)
return 1;

/* permutation check */
memset(tp1, 0, sizeof(tp1));
for(i=0; i<permsize; ++i)
if(++tp1[p[i]] > 1)
return 2;

remap_perm(p);
return 0; /* ok */
} /* line_to_perm */

/* a series of output routines; most are only used for debugging purposes.
 * The first prints a string, a level number, and a permutation */
disp_perm(string, p, n)
char *string;
short *p;
{
int i;
if(remap_mode && !opt_q) {
apply(idn->p, remap, tp1, 1);
apply(tp1, p, tp2, 0);
apply(tp2, remap, tp1, 0);
} else permcpy(tp1, p);
printf("%d@%s ", n, string);
for(i=0; i<permsize; ++i)
printf(" %d", tp1[i]);
nl();
} /* disp_perm */

/* the path back to start */
static char path_p[PATHLEN];
static short path_e[PATHLEN];
static short cyclen[NUM_MOVES];

/* append to the array that holds the path back to start.
 * If perm 3 has a cycle length of 4, then reduce 3^4 to 0.
 * If perms 3 and 2 commute, then reduce 2*3^2*2 to 2^2*3^2.
 * More could be done to reduce paths, if you have
 * a commutes with b and c, but b and c do not commute.
 * Since this situation doesn't arrise in the Rubik puzzles,
 * I'm not going to worry about it. */
buildpath(permnum, direction)
{
int l, mod, i;

for(i=pathlen-1; i>=0; --i) {
if(path_p[i] == permnum) {
l = path_e[i] + direction;
if(mod = cyclen[permnum]) {
if(l > mod/2) l -= mod;
if(l <= -((mod+1)/2)) l += mod;
}
path_e[i] = l;
if(!l) {
--pathlen;
for(; i<pathlen; ++i)
path_p[i] = path_p[i+1], path_e[i] = path_e[i+1];
}
return;
}

if(!theycommute(moveperm[permnum], moveperm[path_p[i]])) break;
}

if(pathlen == PATHLEN)
eprint("1path to start is too long, limit %d moves", PATHLEN);
if(direction == -1 && cyclen[permnum] == 2)
direction = 1;
path_p[pathlen] = permnum;
path_e[pathlen] = direction;
++pathlen;
} /* buildpath */

/* print out the path to start */
printaccess()
{
int i, j;
char c;

for(i=0; i<pathlen; ++i) {
c = movename[path_p[i]];
j = path_e[i];
if(j < 0) j = -j, c = tolower(c);
while(j--) putchar(c);
}
putchar('\n');
} /* printaccess */

/* access this permutation using given generators */
showaccess(c, invert)
struct perm *c;
{
struct connect *chain;
if(!invert)
LOOP_connect(&c->produce, chain)
if(!chain->strong)
buildpath(chain->given, 1-2*chain->invert);
else showaccess(chain->strong, chain->invert);
END_LOOP_connect
else
BACK_LOOP_connect(&c->produce, chain)
if(!chain->strong)
buildpath(chain->given, 2*chain->invert-1);
else showaccess(chain->strong, chain->invert ^ 1);
END_LOOP_connect
} /* showaccess */

/* check path back to start.
 * Called with the entered configuration, which is reduced back to start.
 * This should only be called immediately after showaccess,
 * so that the paht is set up. */
checkpath(icfg)
short *icfg;
{
int i, j, dir;

for(i=0; i<pathlen; ++i) {
dir = 0; /* don't invert */
j = path_e[i];
if(j<0) j = -j, dir = 1;
while(j--) {
apply(icfg, moveperm[path_p[i]], tp1, dir);
permcpy(icfg, tp1);
}
}
if(memcmp(idn->p, icfg, PERMLEN))
eprint("@path back to start is not correct");
} /* checkpath */

showtrail(lc)
struct list_connect *lc;
{
struct connect *c;
int n;
LOOP_connect(lc, c)
printf("%d.%d",c->given,c->invert);
if(c->strong) {
for(n=0; n<permsize; ++n)
if(c->strong->p[n] != n) break;
printf("stronG[%d>%d]",n,c->strong->p[n]);
}
printf(",");
END_LOOP_connect
nl();
} /* showtrail */

/* print the order of the group, and the average path length back to start */
print_order()
{
char first;
int i, n;
long l, lw, avg, worst;
struct perm *p;

first = 1;
printf("the order of the group is ");
for(n=0; n<permsize; ++n)
if(cosrep[n].n_nodes) {
if(!first) printf("*");
first = 0;
printf("%d", cosrep[n].n_nodes+1);
}
if(first) printf("1");
nl();

avg = worst = 0;
for(n=0; n<permsize; ++n) {
l = lw = 0;
LOOP_perm(&cosrep[n], p)
l += p->tostart;
if(p->tostart > lw) lw = p->tostart;
END_LOOP_perm
worst += lw;
avg += l/(cosrep[n].n_nodes + 1);
}
printf("%d rotations specified\n", num_rots);
if(!opt_q) printf("Average path back to start is %ld, worst path back to start is %ld.\n", avg, worst);
} /* print_order */

/* print the order of the group, but show it as one number, if possible */
printord1()
{
int n, x;
struct perm *p;
long ordnum = 1;
char first = 1;

for(n=0; n<permsize; ++n) {
x = 1 + cosrep[n].n_nodes;
if(0x7fffff00/x < ordnum) {
ordnum = 0;
break;
}
ordnum *= x;
}

if(ordnum) { printf("%ld", ordnum); return; }

for(n=0; n<permsize; ++n) {
x = 1 + cosrep[n].n_nodes;
if(x == 1) continue;
if(!first) printf("*");
first = 0;
printf("%d", x);
}
} /* printord1 */

/* dump the generators */
dump()
{
int n;
struct perm *c;
char atrtext[5];

printf("\ndump of strong generators\n\n");
for(n=0; n<PERMSIZE; ++n) {
if(!cosrep[n].n_nodes) continue;
printf("%d reps at level %d\n", cosrep[n].n_nodes+1, n);
LOOP_perm(&cosrep[n], c)
sprintf(atrtext, "%x", c->attrib);
disp_perm(atrtext, c->p, n);
if(!opt_q) {
pathlen = 0;
showaccess(c, 0);
printaccess();
}
END_LOOP_perm
}
} /* dump */

/* end output routines, now for some permutation utilities */

/* do two permutations commute? */
theycommute(t1, t2)
short * t1, *t2;
{
int i;

for(i=0; i<permsize; ++i)
if(t1[t2[i]] != t2[t1[i]]) return 0;
return 1;
} /* theycomute */

/* compose permutations */
apply(a, b, c, invert)
short *a, *b, *c;
int invert; /* invert the second permutation? */
{
register short i;
static short inv[PERMSIZE]; /* to invert b */

if(!invert)
for(i=0; i<permsize; ++i) c[i] = b[a[i]];
else { /* invert b */
for(i=0; i<permsize; ++i) inv[b[i]] = i;
for(i=0; i<permsize; ++i) c[i] = inv[a[i]];
}
} /* apply */

/* find the cycle length of the given permutation */
get_cyclen(p)
short *p;
{
int i, j, start, len, l;

l = 1;
for(i=0; i<permsize; ++i) {
if(p[i] == i) continue;
start = i;
len = 1;
for(j=p[i]; j!=start; j = p[j])
++len;
l = lcm_with(l, len);
if(l == 0) break;
}
return l;
} /* get_cyclen */

/* take least common multiple with next cycle length */
static lcm_with(a1, b1)
{
int a = a1, b = b1;
int c;
long l;

while(b) {
a = a % b;
c = a, a = b, b = c;
}
b1 /= a;
l = a1*b1;
if(l > 30000) l = 0;
return l;
} /* lcm_with */

cyclen_set(permnum)
{
cyclen[permnum] = get_cyclen(moveperm[permnum]);
} /* cyclen_set */

/* transitive closure on rotations */
rots_closure()
{
int base_rots = num_rots;
int i, j, k;
char change;

change = 1;
while(change) {
change = 0;
for(i=0; i<num_rots; ++i) {
for(j=0; j<base_rots; ++j) {
apply(rots[i], rots[j], tp1, 0);
/* have we seen this one before? */
for(k=0; k<num_rots; ++k)
if(!memcmp(tp1, rots[k], PERMLEN)) break;
if(k < num_rots) continue;
if(num_rots == NUM_MOVES)
eprint("1too many rotations, limit %d\n", NUM_MOVES);
rots[k] = (short*) emalloc(PERMLEN);
permcpy(rots[k], tp1);
++num_rots;
change = 1;
} /* loop over base rotations */
} /* loop over all rotatios */
} /* loop while adding new rotations */
} /* rots_closure */

/* each move in the path should map to another move under the rotation */
rotate_moves(rot)
{
int i, j;

for(i=0; i<pathlen; ++i) {
for(j=0; j<num_moves; ++j) {
apply(rots[rot], moveperm[j], tp1, 0);
apply(tp1, rots[rot], tp2, 1);
if(!memcmp(tp2, moveperm[path_p[i]], PERMLEN)) {
path_p[i] = j;
break;
}
/* one move might map to the inverse of the other */
apply(tp2, moveperm[path_p[i]], tp1, 0);
if(!memcmp(tp1, idn->p, PERMLEN)) {
path_p[i] = j;
if(path_e[i] * 2 != cyclen[j])
path_e[i] = -path_e[i];
break;
}
}
if(j == num_moves)
eprint("@some moves have no counterparts under the specified rotations");
}
} /* rotate_moves */

/* what level does this permutation belong in? */
find_level(p)
short *p;
{
int n;
for(n=0; n<permsize; ++n)
if(p[n] != n) break;
return n;
} /* find_level */

/* return a pointer to the coset rep at level n that moves n to dest */
struct perm *find_cosrep(n, dest, mask, force)
short mask; /* all bits in mask&attrib must be true */
int force; /* force this rep to exist */
{
struct perm *p;

if(n == dest)
eprint("@looking for identity coset rep");

LOOP_perm(&cosrep[n], p)
if(p->p[n] != dest) continue;
if(!(p->attrib&CR_INGROUP))
eprint("@level %d rep for coset %d found, but not part of the larger group", n, dest);
if((p->attrib&mask) != mask) continue;
return p;
END_LOOP_perm

if(force)
eprint("@no coset rep at level %d for coset %d\n", n, dest);
return 0; /* not found */
} /* find_cosrep */

